/*
 * Common display-related functions.
 *
 * Copyright (C) 2018 Paul Boddie <paul@boddie.org.uk>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "display.h"



/* Initialise a display configuration. */

void init_display(display_config_t *cfg, uint8_t *framebuffer,
                  uint32_t line_length, uint32_t line_count, int frames)
{
    /* Framebuffer address. */

    cfg->framebuffer = framebuffer;

    /* Frame allocation limits. */

    cfg->total_lines = (line_count + 1) * frames;
    cfg->max_frames = frames;

    /* Screen size and dimensions. */

    cfg->screen_size = line_length * line_count;
    cfg->line_length = line_length;
    cfg->line_count = line_count;

    /* Set the number of frames and the current frame. */

    cfg->frames = frames;
    cfg->frame = 0;

    init_display_properties(cfg, 0);
}

/* Initialise the display constraints. */

void init_display_properties(display_config_t *cfg, uint32_t offset)
{
    /* Fixed address of the frame. */

    cfg->frame_start = cfg->framebuffer +
                       (cfg->screen_size + cfg->line_length) * cfg->frame;

    /* Floating address of the screen contents. */

    cfg->screen_start = cfg->frame_start + offset;

    /* Fixed limit of the frame. */

    cfg->screen_limit = cfg->frame_start + cfg->screen_size;

    /* Recalculate the line multiplier. */

    cfg->line_multiplier = cfg->scanlines / cfg->line_count;
}

/* Set the number of frames in the framebuffer memory. */

void set_frames(display_config_t *cfg, int frames)
{
    if ((frames <= 0) || (frames > cfg->max_frames))
        return;

    /* Recalculate the number of lines. */

    cfg->line_count = (cfg->total_lines - cfg->max_frames) / frames;

    /* Recalculate the screen size. */

    cfg->screen_size = cfg->line_count * cfg->line_length;

    /* Set the number of frames and the current frame. */

    cfg->frames = frames;
    cfg->frame = 0;

    init_display_properties(cfg, 0);
}

/* Select a frame in the framebuffer. */

void select_frame(display_config_t *cfg, int frame, uint32_t offset)
{
    if ((frame < 0) || (frame >= cfg->frames))
        return;

    /* Update the frame details. */

    cfg->frame = frame;

    /* Set the screen start offset when switching frames. */

    init_display_properties(cfg, offset);
}

/* Return the screen start offset. */

uint32_t get_start_offset(display_config_t *cfg)
{
    return cfg->screen_start - cfg->frame_start;
}

/* Return the line data position for the given pixel. */

int get_position(display_config_t *cfg, int x)
{
    int cell, offset, pos;

    if (cfg->line_channels < 2)
        return x;

    /* Determine which cell is providing the position and the offset of the
       pixel within the cell. */

    cell = x / cfg->cell_size;
    offset = x % cfg->cell_size;

    /* Determine the resulting position within the divided-up data. */

    pos = (cell / 2) * cfg->cell_size + offset;

    /* Return the final position within the entire data. All cells in
       odd-numbered positions occur in the first half, all even-numbered cells
       in the second half. */

    return cell % 2 ? pos + cfg->line_length / 2 : pos;
}

/* Provide a pattern to test the line data. */

void test_linedata(display_config_t *cfg)
{
    int x, y;
    uint8_t *linedata = cfg->screen_start;

    for (y = 0; y < cfg->line_count; y++)
    {
        for (x = 0; x < cfg->line_length; x++)
        {
            /* Pixel: I0RRGGBB = Y0YYYYXX */

            linedata[get_position(cfg, x)] = (x % 2) ?
                    (((y / (cfg->line_count / 32)) & 0b1) << 7) |
                    (((y / (cfg->line_count / 16)) & 0b1111) << 2) |
                    ((x / (cfg->line_length / 4)) & 0b11) :
                0x00;
        }

        linedata = wrap_screen_pointer(cfg, linedata + cfg->line_length);
    }
}

/* Copying from/to the display to/from a backing store. */

void copy_display(display_config_t *cfg, uint8_t *store,
                  int width, int height, int ystep,
                  int x, int y, int key, int to_display)
{
    copy_display_section(cfg, store, width, height,
                         0, 0, width, height, ystep,
                         x, y, key, to_display);
}

/* Copying from/to the display to/from a backing store region. */

void copy_display_section(display_config_t *cfg, uint8_t *store,
                          int width, int height,
                          int xstart, int ystart, int xsize, int ysize, int ystep,
                          int x, int y, int key, int to_display)
{
    int sx, sy, dx, dy;
    uint8_t *storeline = store + ystart * width,
            *displayline = wrap_screen_pointer(cfg, cfg->screen_start + y * cfg->line_length),
            pixel;

    /* Define the limits of the copying in the store. */

    int xlimit = xstart + xsize, ylimit = ystart + ysize;

    if (xlimit > width)
        xlimit = width;

    if (ylimit > height)
        ylimit = height;

    /* Perform the copying between the store and display. */

    for (sy = ystart, dy = y; (sy < ylimit) && (dy < cfg->line_count); sy += ystep, dy++)
    {
        for (sx = xstart, dx = x; (sx < xlimit) && (dx < cfg->line_length); sx++, dx++)
        {
            if (to_display)
            {
                pixel = storeline[sx];
                if ((key < 0) || (pixel != key))
                    displayline[get_position(cfg, dx)] = pixel;
            }
            else
                storeline[sx] = displayline[get_position(cfg, dx)];
        }

        storeline += width * ystep;
        displayline = wrap_screen_pointer(cfg, displayline + cfg->line_length);
    }
}

/* Scroll the display. */

void scroll_display(display_config_t *cfg, int x, int y)
{
    /* Move the screen start by the given number of bytes and lines, wrapping
       around the start and end of the framebuffer. */

    cfg->screen_start = wrap_screen_pointer(cfg, cfg->screen_start + x +
                                                 y * cfg->line_length);
}

/* Wrap a pointer within the given limits. */

uint8_t *wrap_pointer(uint8_t *ptr, uint8_t *lower, uint8_t *upper)
{
    uint32_t size = upper - lower;

    if (ptr < lower)
        return upper - (lower - ptr) % size;
    else if (ptr >= upper)
        return lower + (ptr - upper) % size;
    else
        return ptr;
}

/* Wrap the screen pointer to the current frame. */

uint8_t *wrap_screen_pointer(display_config_t *cfg, uint8_t *ptr)
{
    return wrap_pointer(ptr, cfg->frame_start, cfg->screen_limit);
}
